/* 
 * This example uses Virtual filesystem (VFS) component which provides a unified 
 * interface for drivers which can perform operations on file-like objects.
 * 
 * The intend is be very compatible with official SD and SD_MMC libraries.
 *
 * Author: Evandro Luis Copercini
 * based on SD_Test.ino from EspressIf 
 * 
*/
#include <WiFi.h>
#include <WebServer.h>
#include <FS.h>
WebServer server(80);
#include <SPIFFS.h>
File fsUploadFile;
size_t total,used;

char* ssid_ap = "ESP32_SPIFFS";
char* pass_ap = "1234567890";
int val;

//format bytes
String formatBytes(size_t bytes){
  if (bytes < 1024){
    return String(bytes)+"B";
  } else if(bytes < (1024 * 1024)){
    return String(bytes/1024.0)+"KB";
  } else if(bytes < (1024 * 1024 * 1024)){
    return String(bytes/1024.0/1024.0)+"MB";
  } else {
    return String(bytes/1024.0/1024.0/1024.0)+"GB";
  }
}

String getContentType(String filename){
  if(server.hasArg("download")) return "application/octet-stream";
  else if(filename.endsWith(".htm")) return "text/html";
  else if(filename.endsWith(".html")) return "text/html";
  else if(filename.endsWith(".css")) return "text/css";
  else if(filename.endsWith(".js")) return "application/javascript";
  else if(filename.endsWith(".png")) return "image/png";
  else if(filename.endsWith(".gif")) return "image/gif";
  else if(filename.endsWith(".jpg")) return "image/jpeg";
  else if(filename.endsWith(".ico")) return "image/x-icon";
  else if(filename.endsWith(".xml")) return "text/xml";
  else if(filename.endsWith(".pdf")) return "application/x-pdf";
  else if(filename.endsWith(".zip")) return "application/x-zip";
  else if(filename.endsWith(".gz")) return "application/x-gzip";
  return "text/plain";
}

bool handleFileRead(String path){
  Serial.println("handleFileRead: " + path);
  if(path.endsWith("/")) path += "index.htm";
  String contentType = getContentType(path);
  String pathWithGz = path + ".gz";
  if(SPIFFS.exists(pathWithGz) || SPIFFS.exists(path)){
    if(SPIFFS.exists(pathWithGz)) 
      path += ".gz";
    File file = SPIFFS.open(path, "r");
    server.streamFile(file, contentType);
    file.close();
    return true;
  }
  return false;
}

void handleFileUpload(){
  if(server.uri() != "/edit") return;
  HTTPUpload& upload = server.upload();
  if(upload.status == UPLOAD_FILE_START){
    String filename = upload.filename;
    if(!filename.startsWith("/")) filename = "/"+filename;
    fsUploadFile = SPIFFS.open(filename, "w");
    filename = String();
  } else if(upload.status == UPLOAD_FILE_WRITE){
    if(fsUploadFile)
      fsUploadFile.write(upload.buf, upload.currentSize);
  } else if(upload.status == UPLOAD_FILE_END){
    if(fsUploadFile)
      fsUploadFile.close();
  }
}

void handleFileDelete(){
  if(server.args() == 0) return server.send(500, "text/plain", "BAD ARGS");
  String path = server.arg(0);
  if(path == "/")
    return server.send(500, "text/plain", "BAD PATH");
  if(!SPIFFS.exists(path))
    return server.send(404, "text/plain", "FileNotFound");
  SPIFFS.remove(path);
  server.send(200, "text/plain", "");
  path = String();
}

void handleFileCreate(){
  if(server.args() == 0)
    return server.send(500, "text/plain", "BAD ARGS");
  String path = server.arg(0);
  if(path == "/")
    return server.send(500, "text/plain", "BAD PATH");
  if(SPIFFS.exists(path))
    return server.send(500, "text/plain", "FILE EXISTS");
  File file = SPIFFS.open(path, "w");
  if(file)
    file.close();
  else
    return server.send(500, "text/plain", "CREATE FAILED");
  server.send(200, "text/plain", "");
  path = String();
}

void handleFileList() {
  if(!server.hasArg("dir")) {server.send(500, "text/plain", "BAD ARGS"); return;}
  String path = server.arg("dir");
  if(path != "/" && !SPIFFS.exists((char *)path.c_str())) {server.send(500, "text/plain", "BAD PATH"); return;}
  File dir = SPIFFS.open((char *)path.c_str());
  path = String();
  if(!dir.isDirectory()){
    dir.close();
    server.send(500, "text/plain", "BAD DIR");
    return;
  }
  dir.rewindDirectory();
  server.setContentLength(CONTENT_LENGTH_UNKNOWN);
  server.send(200, "text/json", "");

  server.sendContent("[");
  for (int cnt = 0; true; ++cnt) {
    File entry = dir.openNextFile();
    if (!entry)
    break;

    String output;
    if (cnt > 0)
      output = ',';
    output += "{\"type\":\"";
    output += (entry.isDirectory()) ? "dir" : "file";
    output += "\",\"name\":\"";
    output += String(entry.name()).substring(1);
    output += "\"";
    output += "}";
    server.sendContent(output);
    entry.close();
    Serial.println(output);
 }
 server.sendContent("]");
 dir.close();
}

void info() {
  total = SPIFFS.totalBytes();
  used = SPIFFS.usedBytes();
  Serial.println("*** SPIFFS info ***");
  Serial.print("SPIFFS size: "); Serial.print(total); Serial.println(" B");
  Serial.print("SPIFFS used: "); Serial.print(used); Serial.println(" B");
  Serial.print("SPIFFS free: "); Serial.print(total - used); Serial.println(" B");
  Serial.println("");
}

void setup(void) {
  Serial.begin(115200);
  Serial.println("");
  Serial.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
  Serial.println("$    Command 'I' in terminal arduino IDE -> info SPIFFS          $");
  Serial.println("$    Command 'F' in terminal arduino IDE -> formatting SPIFFS    $");
  Serial.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
  Serial.println("");
  SPIFFS.begin(true);

  WiFi.mode(WIFI_AP);   // Access Point mode
  WiFi.softAP(ssid_ap, pass_ap);

  server.on("/list", HTTP_GET, handleFileList);
  server.on("/edit", HTTP_GET, []() {if(!handleFileRead("/edit.htm")) server.send(404, "text/plain", "FileNotFound");} );
  server.on("/edit", HTTP_PUT, handleFileCreate);
  server.on("/edit", HTTP_DELETE, handleFileDelete);
  server.on("/edit", HTTP_POST, [](){ server.send(200, "text/plain", ""); }, handleFileUpload);
  server.onNotFound([](){ if(!handleFileRead(server.uri())) server.send(404, "text/plain", "FileNotFound"); });
  
  server.begin();
  info();
}

void loop(void) {
  server.handleClient();
    val = Serial.read (); // консоль IDE Arduino
  if (val == 'I') {
    info();
  }
  if (val == 'F') { 
    Serial.println("*** Formatting SPIFFS ***");
    SPIFFS.format();
    Serial.println("*** Formatting completed! ***");
    Serial.println("");
    info();
    }
}


